@@ -1,9 +1,10 @@
+import logging
 from dataclasses import dataclass, fields
 from enum import Enum
-from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
-import logging
+from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, cast
 
-from singer.catalog import Catalog as BaseCatalog, CatalogEntry as BaseCatalogEntry
+from singer.catalog import Catalog as BaseCatalog
+from singer.catalog import CatalogEntry as BaseCatalogEntry
 from singer.schema import Schema
 
 Breadcrumb = Tuple[str, ...]
@@ -47,13 +48,13 @@ class Metadata:
         """Parse metadata dictionary."""
         return cls(
             **{
-                field.name.replace("-", "_"): value.get(field.name)
+                field.name: value.get(field.name.replace("_", "-"))
                 for field in fields(cls)
             }
         )
 
     def to_dict(self) -> Dict[str, Any]:
-        """Convert metadata to a JSON-encodable dictionary."""
+        """Convert metadata to a JSON-encodeable dictionary."""
         result = {}
 
         for field in fields(self):
@@ -131,7 +132,7 @@ class MetadataMapping(Dict[Breadcrumb, Union[Metadata, StreamMetadata]]):
             if schema_name:
                 root.schema_name = schema_name
 
-            for field_name in schema["properties"].keys():
+            for field_name in schema.get("properties", {}).keys():
                 if key_properties and field_name in key_properties:
                     entry = Metadata(inclusion=Metadata.InclusionType.AUTOMATIC)
                 else:
@@ -256,6 +257,14 @@ class Catalog(Dict[str, CatalogEntry], BaseCatalog):
             instance[entry.tap_stream_id] = entry
         return instance
 
+    def to_dict(self) -> Dict[str, Any]:
+        """Return a dictionary representation of the catalog.
+
+        Returns:
+            A dictionary with the defined catalog streams.
+        """
+        return cast(Dict[str, Any], super().to_dict())
+
     @property
     def streams(self) -> List[CatalogEntry]:
         """Get catalog entries."""
