@@ -7,12 +7,18 @@ import abc
 import copy
 import hashlib
 import logging
-from typing import Dict, Callable, Any, Tuple, Type, Union, cast, Optional, List
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
 
-from singer_sdk.helpers._singer import Catalog
-from singer_sdk.helpers._catalog import get_selected_schema
-from singer_sdk.helpers import _simpleeval as simpleeval
 from singer_sdk.exceptions import MapExpressionError, StreamMapConfigError
+from singer_sdk.helpers import _simpleeval as simpleeval
+from singer_sdk.helpers._catalog import get_selected_schema
+from singer_sdk.helpers._flattening import (
+    FlatteningOptions,
+    flatten_record,
+    flatten_schema,
+    get_flattening_options,
+)
+from singer_sdk.helpers._singer import Catalog
 from singer_sdk.typing import (
     CustomType,
     IntegerType,
@@ -31,7 +37,14 @@ MAPPER_KEY_PROPERTIES_OPTION = "__key_properties__"
 
 
 def md5(input: str) -> str:
-    """Return the md5 as a string. This is a function for inline calculations."""
+    """Digest a string using MD5. This is a function for inline calculations.
+
+    Args:
+        input: String to digest.
+
+    Returns:
+        A string digested into MD5.
+    """
     return hashlib.md5(input.encode("utf-8")).hexdigest()
 
 
@@ -39,23 +52,111 @@ class StreamMap(metaclass=abc.ABCMeta):
     """Abstract base class for all map classes."""
 
     def __init__(
-        self, stream_alias: str, raw_schema: dict, key_properties: Optional[List[str]]
+        self,
+        stream_alias: str,
+        raw_schema: dict,
+        key_properties: Optional[List[str]],
+        flattening_options: Optional[FlatteningOptions],
     ) -> None:
-        """Initialize mapper."""
+        """Initialize mapper.
+
+        Args:
+            stream_alias: Stream name.
+            raw_schema: Original stream JSON schema.
+            key_properties: Primary key of the source stream.
+            flattening_options: Flattening options, or None to skip flattening.
+        """
         self.stream_alias = stream_alias
         self.raw_schema = raw_schema
         self.raw_key_properties = key_properties
         self.transformed_schema = raw_schema
         self.transformed_key_properties = key_properties
+        self.flattening_options = flattening_options
+        if self.flattening_enabled:
+            self.transformed_schema = self.flatten_schema(self.transformed_schema)
+
+    @property
+    def flattening_enabled(self) -> bool:
+        """True if flattening is enabled for this stream map.
+
+        Returns:
+            True if flattening is enabled, otherwise False.
+        """
+        return (
+            self.flattening_options is not None
+            and self.flattening_options.flattening_enabled
+            and self.flattening_options.max_level > 0
+        )
+
+    def flatten_record(self, record: dict) -> dict:
+        """If flattening is enabled, flatten a record and return the result.
+
+        If flattening is disabled, the original record will be returned.
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            A new dictionary representing the flattened record.
+        """
+        if not self.flattening_options or not self.flattening_enabled:
+            return record
+
+        return flatten_record(
+            record,
+            flattened_schema=self.transformed_schema,
+            max_level=self.flattening_options.max_level,
+            separator=self.flattening_options.separator,
+        )
+
+    def flatten_schema(self, raw_schema: dict) -> dict:
+        """Flatten the provided schema.
+
+        Args:
+            raw_schema: The raw schema to flatten.
+
+        Returns:
+            The flattened version of the schema.
+        """
+        if not self.flattening_options or not self.flattening_enabled:
+            return raw_schema
+
+        return flatten_schema(
+            raw_schema,
+            separator=self.flattening_options.separator,
+            max_level=self.flattening_options.max_level,
+        )
 
     @abc.abstractmethod
     def transform(self, record: dict) -> Optional[dict]:
-        """Transform a record and return the result."""
-        raise NotImplementedError
+        """Transform a record and return the result.
+
+        Record flattening will also be performed, if enabled.
+
+        Subclasses should call the super().transform(record) after any other custom
+        transforms are performed.
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            A new dictionary representing a transformed record.
+        """
+        return self.flatten_record(record)
 
     @abc.abstractmethod
     def get_filter_result(self, record: dict) -> bool:
-        """Return True to include the record or False to exclude."""
+        """Exclude records from a stream.
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Return:
+            True to include the record or False to exclude.
+
+        Raises:
+            NotImplementedError: If the derived class doesn't override this method.
+        """
         raise NotImplementedError
 
 
@@ -66,25 +167,53 @@ class DefaultStreamMap(StreamMap):
 class RemoveRecordTransform(DefaultStreamMap):
     """Default mapper which simply excludes any records."""
 
-    def transform(self, record: dict) -> Optional[dict]:
-        """Return None (always exclude)."""
+    def transform(self, record: dict) -> None:
+        """Return None (always exclude).
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            None
+        """
         _ = record  # Drop the record
         return None
 
     def get_filter_result(self, record: dict) -> bool:
-        """Return False (always exclude)."""
+        """Exclude all records.
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            Always `False`.
+        """
         return False
 
 
 class SameRecordTransform(DefaultStreamMap):
     """Default mapper which simply returns the original records."""
 
-    def transform(self, record: dict) -> Optional[dict]:
-        """Return original record unchanged."""
-        return record
+    def transform(self, record: dict) -> dict:
+        """Return original record unchanged.
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            The original record unchanged.
+        """
+        return cast(dict, super().transform(record))
 
     def get_filter_result(self, record: dict) -> bool:
-        """Return True (always include)."""
+        """Return True (always include).
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            Always `True`.
+        """
         return True
 
 
@@ -98,12 +227,23 @@ class CustomStreamMap(StreamMap):
         raw_schema: dict,
         key_properties: Optional[List[str]],
         map_transform: dict,
+        flattening_options: Optional[FlatteningOptions],
     ) -> None:
-        """Initialize mapper."""
+        """Initialize mapper.
+
+        Args:
+            stream_alias: Stream name.
+            map_config: Stream map configuration.
+            raw_schema: Original stream's JSON schema.
+            key_properties: Primary key of the source stream.
+            map_transform: Dictionary of transformations to apply to the stream.
+            flattening_options: Flattening options, or None to skip flattening.
+        """
         super().__init__(
             stream_alias=stream_alias,
             raw_schema=raw_schema,
             key_properties=key_properties,
+            flattening_options=flattening_options,
         )
 
         self.map_config = map_config
@@ -116,22 +256,56 @@ class CustomStreamMap(StreamMap):
         ) = self._init_functions_and_schema(stream_map=map_transform)
 
     def transform(self, record: dict) -> Optional[dict]:
-        """Return a transformed record."""
-        return self._transform_fn(record)
+        """Return a transformed record.
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            The transformed record.
+        """
+        transformed_record = self._transform_fn(record)
+        if not transformed_record:
+            return None
+
+        return super().transform(transformed_record)
 
     def get_filter_result(self, record: dict) -> bool:
-        """Return True to include or False to exclude."""
+        """Return True to include or False to exclude.
+
+        Args:
+            record: An individual record dictionary in a stream.
+
+        Returns:
+            Boolean flag for record selection.
+        """
         return self._filter_fn(record)
 
     @property
     def functions(self) -> Dict[str, Callable]:
-        """Return the functions which should be available for expression evaluation."""
+        """Get availabale transformation functions.
+
+        Returns:
+            Functions which should be available for expression evaluation.
+        """
         funcs: Dict[str, Any] = simpleeval.DEFAULT_FUNCTIONS.copy()
         funcs["md5"] = md5
         return funcs
 
     def _eval(self, expr: str, record: dict, property_name: Optional[str]) -> Any:
-        """Solve an expression."""
+        """Solve an expression.
+
+        Args:
+            expr: String expression to evaluate.
+            record: Individual stream record.
+            property_name: Name of property to transform in the record.
+
+        Returns:
+            Evaluated expression.
+
+        Raises:
+            MapExpressionError: If the mapping expression failed to evaluate.
+        """
         names = record.copy()  # Start with names from record properties
         names["_"] = record  # Add a shorthand alias in case of reserved words in names
         names["record"] = record  # ...and a longhand alias
@@ -151,7 +325,15 @@ class CustomStreamMap(StreamMap):
     def _eval_type(
         self, expr: str, default: Optional[JSONTypeHelper] = None
     ) -> JSONTypeHelper:
-        """Evaluate an expression's type."""
+        """Evaluate an expression's type.
+
+        Args:
+            expr: String expression to evaluate.
+            default: TODO.
+
+        Returns:
+            TODO
+        """
         assert expr is not None, "Expression should be str, not None"
 
         default = default or StringType()
@@ -173,7 +355,18 @@ class CustomStreamMap(StreamMap):
     def _init_functions_and_schema(
         self, stream_map: dict
     ) -> Tuple[Callable[[dict], bool], Callable[[dict], Optional[dict]], dict]:
-        """Return a tuple: filter_fn, transform_fn, transformed_schema."""
+        """Return a tuple: filter_fn, transform_fn, transformed_schema.
+
+        Args:
+            stream_map: TODO
+
+        Returns:
+            TODO.
+
+        Raises:
+            NotImplementedError: TODO
+            StreamMapConfigError: TODO
+        """
         stream_map = copy.copy(stream_map)
 
         filter_rule: Optional[str] = None
@@ -183,7 +376,7 @@ class CustomStreamMap(StreamMap):
             logging.info(f"Found '{self.stream_alias}' filter rule: {filter_rule}")
 
         if stream_map and MAPPER_KEY_PROPERTIES_OPTION in stream_map:
-            self.transformed_key_properties = stream_map.pop(
+            self.transformed_key_properties: List[str] = stream_map.pop(
                 MAPPER_KEY_PROPERTIES_OPTION
             )
             logging.info(
@@ -211,10 +404,6 @@ class CustomStreamMap(StreamMap):
         if not include_by_default:
             # Start with only the defined (or transformed) key properties
             transformed_schema = PropertiesList().to_dict()
-            for key_property in self.transformed_key_properties or []:
-                transformed_schema["properties"][key_property] = self.raw_schema[
-                    "properties"
-                ][key_property]
 
         if "properties" not in transformed_schema:
             transformed_schema["properties"] = {}
@@ -251,6 +440,14 @@ class CustomStreamMap(StreamMap):
                     f"for '{self.stream_alias}:{prop_key}'."
                 )
 
+        for key_property in self.transformed_key_properties or []:
+            if key_property not in transformed_schema["properties"]:
+                raise StreamMapConfigError(
+                    f"Invalid key properties for '{self.stream_alias}': "
+                    f"[{','.join(self.transformed_key_properties)}]. "
+                    f"Property '{key_property}' was not detected in schema."
+                )
+
         # Declare function variables
 
         def eval_filter(record: dict) -> bool:
@@ -295,7 +492,8 @@ class CustomStreamMap(StreamMap):
                 # Start with only the defined (or transformed) key properties
                 result = {}
                 for key_property in self.transformed_key_properties or []:
-                    result[key_property] = record[key_property]
+                    if key_property in record:
+                        result[key_property] = record[key_property]
 
             for prop_key, prop_def in list(stream_map.items()):
                 if prop_def is None:
@@ -327,10 +525,19 @@ class PluginMapper:
         self,
         plugin_config: Dict[str, Dict[str, Union[str, dict]]],
         logger: logging.Logger,
-    ):
-        """Initialize mapper."""
+    ) -> None:
+        """Initialize mapper.
+
+        Args:
+            plugin_config: TODO
+            logger: TODO
+
+        Raises:
+            StreamMapConfigError: TODO
+        """
         self.stream_maps = cast(Dict[str, List[StreamMap]], {})
         self.map_config = plugin_config.get("stream_map_config", {})
+        self.flattening_options = get_flattening_options(plugin_config)
         self.default_mapper_type: Type[DefaultStreamMap] = SameRecordTransform
         self.logger = logger
 
@@ -352,7 +559,7 @@ class PluginMapper:
                 )
         else:
             logging.info(
-                "Operator '{MAPPER_ELSE_OPTION}=None' was not found. "
+                f"Operator '{MAPPER_ELSE_OPTION}=None' was not found. "
                 "Unmapped streams will be included in output."
             )
         for stream_map_key, stream_def in self.stream_maps_dict.items():
@@ -362,7 +569,11 @@ class PluginMapper:
                 )
 
     def register_raw_streams_from_catalog(self, catalog: Catalog) -> None:
-        """Register all streams as described in the catalog dict."""
+        """Register all streams as described in the catalog dict.
+
+        Args:
+            catalog: TODO
+        """
         for catalog_entry in catalog.streams:
             self.register_raw_stream_schema(
                 catalog_entry.stream or catalog_entry.tap_stream_id,
@@ -378,16 +589,43 @@ class PluginMapper:
     def register_raw_stream_schema(
         self, stream_name: str, schema: dict, key_properties: Optional[List[str]]
     ) -> None:
-        """Register a new stream as described by its name and schema."""
+        """Register a new stream as described by its name and schema.
+
+        If stream has already been registered and schema or key_properties has changed,
+        the older registration will be removed and replaced with new, updated mappings.
+
+        Args:
+            stream_name: The stream name.
+            schema: The schema definition for the stream.
+            key_properties: The key properties of the stream.
+
+        Raises:
+            StreamMapConfigError: If the configuration is invalid.
+        """
+        if stream_name in self.stream_maps:
+            primary_mapper = self.stream_maps[stream_name][0]
+            if (
+                primary_mapper.raw_schema != schema
+                or primary_mapper.raw_key_properties != key_properties
+            ):
+                # Unload/reset stream maps if schema or key properties have changed.
+                self.stream_maps.pop(stream_name)
+
         if stream_name not in self.stream_maps:
             # The 0th mapper should be the same-named treatment.
             # Additional items may be added for aliasing or multi projections.
             self.stream_maps[stream_name] = [
-                self.default_mapper_type(stream_name, schema, key_properties)
+                self.default_mapper_type(
+                    stream_name,
+                    schema,
+                    key_properties,
+                    flattening_options=self.flattening_options,
+                )
             ]
 
         for stream_map_key, stream_def in self.stream_maps_dict.items():
             stream_alias: str = stream_map_key
+            source_stream: str = stream_map_key
             if isinstance(stream_def, str):
                 if stream_name == stream_map_key:
                     # TODO: Add any expected cases for str expressions (currently none)
@@ -405,6 +643,7 @@ class PluginMapper:
                     stream_alias=stream_map_key,
                     raw_schema=schema,
                     key_properties=None,
+                    flattening_options=self.flattening_options,
                 )
                 logging.info(f"Set null tansform as default for '{stream_name}'")
                 continue
@@ -416,11 +655,9 @@ class PluginMapper:
                 )
 
             if MAPPER_SOURCE_OPTION in stream_def:
-                if stream_name != cast(str, stream_def.pop(MAPPER_SOURCE_OPTION)):
-                    # Not a match
-                    continue
+                source_stream = cast(str, stream_def.pop(MAPPER_SOURCE_OPTION))
 
-            elif stream_name != stream_map_key:
+            if source_stream != stream_name:
                 # Not a match
                 continue
 
@@ -433,22 +670,13 @@ class PluginMapper:
                 map_config=self.map_config,
                 raw_schema=schema,
                 key_properties=key_properties,
+                flattening_options=self.flattening_options,
             )
-            if stream_name == stream_alias:
-                # Zero-th mapper should be the same-named mapper.
-                # Override the default mapper with this custom map
+
+            if source_stream == stream_map_key:
+                # Zero-th mapper should be the same-keyed mapper.
+                # Override the default mapper with this custom map.
                 self.stream_maps[stream_name][0] = mapper
             else:
                 # Additional mappers for aliasing and multi-projection:
                 self.stream_maps[stream_name].append(mapper)
-
-    def get_primary_mapper(self, stream_name: str) -> StreamMap:
-        """Return the primary mapper for the specified stream name."""
-        if stream_name not in self.stream_maps:
-            raise StreamMapConfigError(f"No mapper found for {stream_name}. ")
-
-        return self.stream_maps[stream_name][0]
-
-    def apply_primary_mapper(self, stream_name: str, record: dict) -> Optional[dict]:
-        """Apply the primary mapper for the stream and return the result."""
-        return self.get_primary_mapper(stream_name).transform(record)
