@@ -5,17 +5,16 @@ import copy
 import datetime
 import json
 import logging
-from singer_sdk.mapper import SameRecordTransform, StreamMap
-import requests
-from types import MappingProxyType
 from os import PathLike
 from pathlib import Path
+from types import MappingProxyType
 from typing import (
+    Any,
     Callable,
     Dict,
-    Any,
-    List,
+    Generator,
     Iterable,
+    List,
     Mapping,
     Optional,
     Tuple,
@@ -26,39 +25,36 @@ from typing import (
 )
 
 import pendulum
+import requests
 import singer
-from singer import (
-    RecordMessage,
-    SchemaMessage,
-    StateMessage,
-)
+from singer import RecordMessage, SchemaMessage, StateMessage
 from singer.schema import Schema
 
-from singer_sdk.plugin_base import PluginBase as TapBaseClass
+from singer_sdk.exceptions import InvalidStreamSortException, MaxRecordsLimitException
 from singer_sdk.helpers._catalog import pop_deselected_record_properties
+from singer_sdk.helpers._compat import final
+from singer_sdk.helpers._flattening import get_flattening_options
 from singer_sdk.helpers._singer import (
     Catalog,
     CatalogEntry,
     MetadataMapping,
     SelectionMask,
 )
-from singer_sdk.helpers._typing import (
-    conform_record_data_types,
-    is_datetime_type,
-)
 from singer_sdk.helpers._state import (
-    get_writeable_state_dict,
-    get_state_partitions_list,
-    increment_state,
     finalize_state_progress_markers,
+    get_starting_replication_value,
+    get_state_partitions_list,
+    get_writeable_state_dict,
+    increment_state,
+    log_sort_error,
     reset_state_progress_markers,
     write_replication_key_signpost,
-    log_sort_error,
+    write_starting_replication_value,
 )
-from singer_sdk.exceptions import MaxRecordsLimitException, InvalidStreamSortException
-from singer_sdk.helpers._compat import final
+from singer_sdk.helpers._typing import conform_record_data_types, is_datetime_type
 from singer_sdk.helpers._util import utc_now
-
+from singer_sdk.mapper import RemoveRecordTransform, SameRecordTransform, StreamMap
+from singer_sdk.plugin_base import PluginBase as TapBaseClass
 
 # Replication methods
 REPLICATION_FULL_TABLE = "FULL_TABLE"
@@ -85,8 +81,18 @@ class Stream(metaclass=abc.ABCMeta):
         tap: TapBaseClass,
         schema: Optional[Union[str, PathLike, Dict[str, Any], Schema]] = None,
         name: Optional[str] = None,
-    ):
-        """Init tap stream."""
+    ) -> None:
+        """Init tap stream.
+
+        Args:
+            tap: Singer Tap this stream belongs to.
+            schema: JSON schema for records in this stream.
+            name: Name of this stream.
+
+        Raises:
+            ValueError: TODO
+            FileNotFoundError: TODO
+        """
         if name:
             self.name: str = name
         if not self.name:
@@ -111,7 +117,7 @@ class Stream(metaclass=abc.ABCMeta):
         if schema:
             if isinstance(schema, (PathLike, str)):
                 if not Path(schema).is_file():
-                    raise FileExistsError(
+                    raise FileNotFoundError(
                         f"Could not find schema file '{self.schema_filepath}'."
                     )
 
@@ -136,9 +142,12 @@ class Stream(metaclass=abc.ABCMeta):
 
     @property
     def stream_maps(self) -> List[StreamMap]:
-        """Return a list of one or more map transformations for this stream.
+        """Get stream transformation maps.
 
         The 0th item is the primary stream map. List should not be empty.
+
+        Returns:
+            A list of one or more map transformations for this stream.
         """
         if self._stream_maps:
             return self._stream_maps
@@ -151,24 +160,28 @@ class Stream(metaclass=abc.ABCMeta):
         else:
             self.logger.info(
                 f"No custom mapper provided for '{self.name}'. "
-                "Using SameRecordTransform as default."
+                "Using SameRecordTransform."
             )
             self._stream_maps = [
                 SameRecordTransform(
                     stream_alias=self.name,
                     raw_schema=self.schema,
                     key_properties=self.primary_keys,
+                    flattening_options=get_flattening_options(self.config),
                 )
             ]
-        return self.stream_maps
+        return self._stream_maps
 
     @property
     def is_timestamp_replication_key(self) -> bool:
-        """Return True if the stream uses a timestamp-based replication key.
+        """Check is replication key is a timestamp.
 
         Developers can override to `True` in order to force this value, although this
         should not be required in most use cases since the type can generally be
         accurately detected from the JSON Schema.
+
+        Returns:
+            True if the stream uses a timestamp-based replication key.
         """
         if not self.replication_key:
             return False
@@ -178,24 +191,29 @@ class Stream(metaclass=abc.ABCMeta):
     def get_starting_replication_key_value(
         self, context: Optional[dict]
     ) -> Optional[Any]:
-        """Return starting replication key.
+        """Get starting replication key.
 
         Will return the value of the stream's replication key when `--state` is passed.
         If no prior state exists, will return `None`.
 
         Developers should use this method to seed incremental processing for
         non-datetime replication keys. For datetime and date replication keys, use
-        `get_starting_timestamp()`
-        """
-        if self.is_timestamp_replication_key:
-            return self.get_starting_timestamp(context)
+        :meth:`~singer_sdk.Stream.get_starting_timestamp()`
 
-        return self._starting_replication_key_value(context)
+        Args:
+            context: Stream partition or context dictionary.
+
+        Returns:
+            Starting replication value.
+        """
+        state = self.get_context_state(context)
+
+        return get_starting_replication_value(state)
 
     def get_starting_timestamp(
         self, context: Optional[dict]
     ) -> Optional[datetime.datetime]:
-        """Return `start_date` config, or state if using timestamp replication.
+        """Get starting replication timestamp.
 
         Will return the value of the stream's replication key when `--state` is passed.
         If no state exists, will return `start_date` if set, or `None` if neither
@@ -203,40 +221,47 @@ class Stream(metaclass=abc.ABCMeta):
 
         Developers should use this method to seed incremental processing for date
         and datetime replication keys. For non-datetime replication keys, use
-        `get_starting_replication_key_value()`
+        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()`
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Returns:
+            `start_date` from config, or state value if using timestamp replication.
+
+        Raises:
+            ValueError: If the replication value is not a valid timestamp.
         """
-        if self.is_timestamp_replication_key:
-            replication_key_value = self._starting_replication_key_value(context)
-            if replication_key_value:
-                return cast(datetime.datetime, pendulum.parse(replication_key_value))
+        value = self.get_starting_replication_key_value(context)
 
-        if "start_date" in self.config:
-            return cast(datetime.datetime, pendulum.parse(self.config["start_date"]))
+        if value is None:
+            return None
 
-        return None
+        if not self.is_timestamp_replication_key:
+            raise ValueError(
+                f"The replication key {self.replication_key} is not of timestamp type"
+            )
 
-    def _starting_replication_key_value(self, context: Optional[dict]) -> Optional[Any]:
-        """Return starting replication key."""
-        if self.replication_key:
-            state = self.get_context_state(context)
-            replication_key_value = state.get("replication_key_value")
-            if replication_key_value and self.replication_key == state.get(
-                "replication_key"
-            ):
-                return replication_key_value
-
-        return None
+        return cast(datetime.datetime, pendulum.parse(value))
 
     @final
     @property
     def selected(self) -> bool:
-        """Return true if the stream is selected."""
+        """Check if stream is selected.
+
+        Returns:
+            True if the stream is selected.
+        """
         return self.mask.get((), True)
 
     @final
     @property
     def has_selected_descendents(self) -> bool:
-        """Return True if any child streams are selected, recursively."""
+        """Check descendents.
+
+        Returns:
+            True if any child streams are selected, recursively.
+        """
         for child in self.child_streams or []:
             if child.selected or child.has_selected_descendents:
                 return True
@@ -246,7 +271,11 @@ class Stream(metaclass=abc.ABCMeta):
     @final
     @property
     def descendent_streams(self) -> List["Stream"]:
-        """Return a list of all children recursively."""
+        """Get child streams.
+
+        Returns:
+            A list of all children, recursively.
+        """
         result: List[Stream] = list(self.child_streams) or []
         for child in self.child_streams:
             result += child.descendent_streams or []
@@ -256,18 +285,47 @@ class Stream(metaclass=abc.ABCMeta):
         self,
         context: Optional[dict],
         value: Union[datetime.datetime, str, int, float],
-    ):
-        """Write the signpost value, if available."""
+    ) -> None:
+        """Write the signpost value, if available.
+
+        Args:
+            context: Stream partition or context dictionary.
+            value: TODO
+
+        Returns:
+            TODO
+        """
         if not value:
             return
 
         state = self.get_context_state(context)
         write_replication_key_signpost(state, value)
 
+    def _write_starting_replication_value(self, context: Optional[dict]) -> None:
+        """Write the starting replication value, if available.
+
+        Args:
+            context: Stream partition or context dictionary.
+        """
+        value = None
+        state = self.get_context_state(context)
+
+        if self.replication_key:
+            replication_key_value = state.get("replication_key_value")
+            if replication_key_value and self.replication_key == state.get(
+                "replication_key"
+            ):
+                value = replication_key_value
+
+            elif "start_date" in self.config:
+                value = self.config["start_date"]
+
+        write_starting_replication_value(state, value)
+
     def get_replication_key_signpost(
         self, context: Optional[dict]
     ) -> Optional[Union[datetime.datetime, Any]]:
-        """Return the max allowable bookmark value for this stream's replication key.
+        """Get the replication signpost.
 
         For timestamp-based replication keys, this defaults to `utc_now()`. For
         non-timestamp replication keys, default to `None`. For consistency in subsequent
@@ -276,7 +334,13 @@ class Stream(metaclass=abc.ABCMeta):
 
         Developers may optionally override this method in advanced use cases such
         as unsorted incremental streams or complex hierarchical stream scenarios.
-        For more info: https://sdk.meltano.com/en/latest/implementation/state.html
+        For more info: :doc:`/implementation/state`
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Returns:
+            Max allowable bookmark value for this stream's replication key.
         """
         if self.is_timestamp_replication_key:
             return utc_now()
@@ -285,32 +349,51 @@ class Stream(metaclass=abc.ABCMeta):
 
     @property
     def schema_filepath(self) -> Optional[Path]:
-        """Return a path to a schema file for the stream or None if n/a."""
+        """Get path to schema file.
+
+        Returns:
+            Path to a schema file for the stream or `None` if n/a.
+        """
         return self._schema_filepath
 
     @property
     def schema(self) -> dict:
-        """Return the schema dict for the stream."""
+        """Get schema.
+
+        Returns:
+            JSON Schema dictionary for this stream.
+        """
         return self._schema
 
     @property
     def primary_keys(self) -> Optional[List[str]]:
-        """Return primary key(s) for the stream."""
+        """Get primary keys.
+
+        Returns:
+            A list of primary key(s) for the stream.
+        """
         if not self._primary_keys:
             return []
         return self._primary_keys
 
     @primary_keys.setter
-    def primary_keys(self, new_value: List[str]):
-        """Set primary key(s) for the stream."""
+    def primary_keys(self, new_value: List[str]) -> None:
+        """Set primary key(s) for the stream.
+
+        Args:
+            new_value: TODO
+        """
         self._primary_keys = new_value
 
     @property
     def state_partitioning_keys(self) -> Optional[List[str]]:
-        """Return partition keys for the stream state bookmarks.
+        """Get state partition keys.
 
         If not set, a default partitioning will be inherited from the stream's context.
         If an empty list is set (`[]`), state will be held in one bookmark per stream.
+
+        Returns:
+            Partition keys for the stream state bookmarks.
         """
         return self._state_partitioning_keys
 
@@ -320,40 +403,57 @@ class Stream(metaclass=abc.ABCMeta):
 
         If not set, a default partitioning will be inherited from the stream's context.
         If an empty list is set (`[]`), state will be held in one bookmark per stream.
+
+        Args:
+            new_value: the new list of keys
         """
         self._state_partitioning_keys = new_value
 
     @property
     def replication_key(self) -> Optional[str]:
-        """Return replication key for the stream."""
+        """Get replication key.
+
+        Returns:
+            Replication key for the stream.
+        """
         if not self._replication_key:
             return None
         return self._replication_key
 
     @replication_key.setter
     def replication_key(self, new_value: str) -> None:
-        """Set replication key for the stream."""
+        """Set replication key for the stream.
+
+        Args:
+            new_value: TODO
+        """
         self._replication_key = new_value
 
     @property
     def is_sorted(self) -> bool:
-        """Return `True` if stream is sorted. Defaults to `False`.
+        """Check if stream is sorted.
 
         When `True`, incremental streams will attempt to resume if unexpectedly
         interrupted.
 
         This setting enables additional checks which may trigger
         `InvalidStreamSortException` if records are found which are unsorted.
+
+        Returns:
+            `True` if stream is sorted. Defaults to `False`.
         """
         return False
 
     @property
     def metadata(self) -> MetadataMapping:
-        """Return a mapping from property breadcrumbs to metadata objects.
+        """Get stream metadata.
 
         Metadata attributes (`inclusion`, `selected`, etc.) are part of the Singer spec.
 
         Metadata from an input catalog will override standard metadata.
+
+        Returns:
+            A mapping from property breadcrumbs to metadata objects.
         """
         if self._metadata is not None:
             return self._metadata
@@ -382,7 +482,11 @@ class Stream(metaclass=abc.ABCMeta):
 
     @property
     def _singer_catalog_entry(self) -> CatalogEntry:
-        """Return catalog entry as specified by the Singer catalog spec."""
+        """Return catalog entry as specified by the Singer catalog spec.
+
+        Returns:
+            TODO
+        """
         return CatalogEntry(
             tap_stream_id=self.tap_stream_id,
             stream=self.name,
@@ -400,11 +504,20 @@ class Stream(metaclass=abc.ABCMeta):
 
     @property
     def _singer_catalog(self) -> Catalog:
+        """TODO.
+
+        Returns:
+            TODO
+        """
         return Catalog([(self.tap_stream_id, self._singer_catalog_entry)])
 
     @property
     def config(self) -> Mapping[str, Any]:
-        """Return a frozen (read-only) config dictionary map."""
+        """Get stream configuration.
+
+        Returns:
+            A frozen (read-only) config dictionary map.
+        """
         return MappingProxyType(self._config)
 
     @property
@@ -413,12 +526,19 @@ class Stream(metaclass=abc.ABCMeta):
 
         Default implementations will return `self.name` but this behavior may be
         overridden if required by the developer.
+
+        Returns:
+            Unique stream ID.
         """
         return self.name
 
     @property
     def replication_method(self) -> str:
-        """Return the replication method to be used."""
+        """Get replication method.
+
+        Returns:
+            Replication method to be used for this stream.
+        """
         if self.forced_replication_method:
             return str(self.forced_replication_method)
         if self.replication_key:
@@ -436,8 +556,13 @@ class Stream(metaclass=abc.ABCMeta):
         This method is internal to the SDK and should not need to be overridden.
         Developers may access this property but this is not recommended except in
         advanced use cases. Instead, developers should access the latest stream
-        replication key values using `Stream.get_starting_timestamp()` for timestamp
-        keys, or `Stream.get_starting_replication_key_value()` for non-timestamp keys.
+        replication key values using :meth:`~singer_sdk.Stream.get_starting_timestamp()`
+        for timestamp keys, or
+        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()` for
+        non-timestamp keys.
+
+        Returns:
+            A writeable state dict for the entire tap.
         """
         return self._tap_state
 
@@ -450,10 +575,20 @@ class Stream(metaclass=abc.ABCMeta):
         This method is internal to the SDK and should not need to be overridden.
         Developers may access this property but this is not recommended except in
         advanced use cases. Instead, developers should access the latest stream
-        replication key values using `Stream.get_starting_timestamp()` for timestamp
-        keys, or `Stream.get_starting_replication_key_value()` for non-timestamp keys.
+        replication key values using
+        :meth:`~singer_sdk.Stream.get_starting_timestamp()` for timestamp keys, or
+        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()` for
+        non-timestamp keys.
 
-        Partition level may be overridden by Stream.state_partitioning_keys if set.
+        Partition level may be overridden by
+        :attr:`~singer_sdk.Stream.state_partitioning_keys` if set.
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Returns:
+            A partitioned context state if applicable; else returns stream state.
+            A blank state will be created in none exists.
         """
         state_partition_context = self._get_state_partition_context(context)
         if state_partition_context:
@@ -466,15 +601,20 @@ class Stream(metaclass=abc.ABCMeta):
 
     @property
     def stream_state(self) -> dict:
-        """Return a writeable state dict for this stream.
+        """Get writable state.
 
         This method is internal to the SDK and should not need to be overridden.
         Developers may access this property but this is not recommended except in
         advanced use cases. Instead, developers should access the latest stream
-        replication key values using `Stream.get_starting_timestamp()` for timestamp
-        keys, or `Stream.get_starting_replication_key_value()` for non-timestamp keys.
+        replication key values using :meth:`~singer_sdk.Stream.get_starting_timestamp()`
+        for timestamp keys, or
+        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()` for
+        non-timestamp keys.
 
         A blank state entry will be created if one doesn't already exist.
+
+        Returns:
+            A writable state dict for this stream.
         """
         return get_writeable_state_dict(self.tap_state, self.name)
 
@@ -482,12 +622,15 @@ class Stream(metaclass=abc.ABCMeta):
 
     @property
     def partitions(self) -> Optional[List[dict]]:
-        """Return a list of partition key dicts (if applicable), otherwise None.
+        """Get stream partitions.
 
         Developers may override this property to provide a default partitions list.
 
         By default, this method returns a list of any partitions which are already
         defined in state, otherwise None.
+
+        Returns:
+            A list of partition key dicts (if applicable), otherwise `None`.
         """
         result: List[dict] = []
         for partition_state in (
@@ -500,11 +643,18 @@ class Stream(metaclass=abc.ABCMeta):
 
     def _increment_stream_state(
         self, latest_record: Dict[str, Any], *, context: Optional[dict] = None
-    ):
+    ) -> None:
         """Update state of stream or partition with data from the provided record.
 
         Raises InvalidStreamSortException is self.is_sorted = True and unsorted data is
         detected.
+
+        Args:
+            latest_record: TODO
+            context: Stream partition or context dictionary.
+
+        Raises:
+            ValueError: TODO
         """
         state_dict = self.get_context_state(context)
         if latest_record:
@@ -530,31 +680,59 @@ class Stream(metaclass=abc.ABCMeta):
 
     # Private message authoring methods:
 
-    def _write_state_message(self):
+    def _write_state_message(self) -> None:
         """Write out a STATE message with the latest state."""
         singer.write_message(StateMessage(value=self.tap_state))
 
-    def _write_schema_message(self):
-        """Write out a SCHEMA message with the stream schema."""
+    def _generate_schema_messages(self) -> Generator[SchemaMessage, None, None]:
+        """Generate schema messages from stream maps.
+
+        Yields:
+            Schema message objects.
+        """
         bookmark_keys = [self.replication_key] if self.replication_key else None
         for stream_map in self.stream_maps:
+            if isinstance(stream_map, RemoveRecordTransform):
+                # Don't emit schema if the stream's records are all ignored.
+                continue
+
             schema_message = SchemaMessage(
                 stream_map.stream_alias,
                 stream_map.transformed_schema,
-                self.primary_keys,
+                stream_map.transformed_key_properties,
                 bookmark_keys,
             )
+            yield schema_message
+
+    def _write_schema_message(self) -> None:
+        """Write out a SCHEMA message with the stream schema."""
+        for schema_message in self._generate_schema_messages():
             singer.write_message(schema_message)
 
     @property
     def mask(self) -> SelectionMask:
-        """Get a boolean mask for stream and property selection."""
+        """Get a boolean mask for stream and property selection.
+
+        Returns:
+            A mapping of breadcrumbs to boolean values, representing stream and field
+            selection.
+        """
         if self._mask is None:
             self._mask = self.metadata.resolve_selection()
         return self._mask
 
-    def _write_record_message(self, record: dict) -> None:
-        """Write out a RECORD message."""
+    def _generate_record_messages(
+        self,
+        record: dict,
+    ) -> Generator[RecordMessage, None, None]:
+        """Write out a RECORD message.
+
+        Args:
+            record: A single stream record.
+
+        Yields:
+            Record message objects.
+        """
         pop_deselected_record_properties(record, self.schema, self.mask, self.logger)
         record = conform_record_data_types(
             stream_name=self.name,
@@ -572,10 +750,28 @@ class Stream(metaclass=abc.ABCMeta):
                     version=None,
                     time_extracted=utc_now(),
                 )
-                singer.write_message(record_message)
+
+                yield record_message
+
+    def _write_record_message(self, record: dict) -> None:
+        """Write out a RECORD message.
+
+        Args:
+            record: A single stream record.
+        """
+        for record_message in self._generate_record_messages(record):
+            singer.write_message(record_message)
 
     @property
     def _metric_logging_function(self) -> Optional[Callable]:
+        """Return the metrics logging function.
+
+        Returns:
+            The logging function for emitting metrics.
+
+        Raises:
+            ValueError: If logging level setting is an unsupported value.
+        """
         if METRICS_LOG_LEVEL_SETTING not in self.config:
             return self.logger.info
 
@@ -588,13 +784,21 @@ class Stream(metaclass=abc.ABCMeta):
         if self.config[METRICS_LOG_LEVEL_SETTING].upper() == "NONE":
             return None
 
-        assert False, (
+        raise ValueError(
             "Unexpected logging level for metrics: "
             + self.config[METRICS_LOG_LEVEL_SETTING]
         )
 
     def _write_metric_log(self, metric: dict, extra_tags: Optional[dict]) -> None:
-        """Emit a metric log. Optionally with appended tag info."""
+        """Emit a metric log. Optionally with appended tag info.
+
+        Args:
+            metric: TODO
+            extra_tags: TODO
+
+        Returns:
+            None
+        """
         if not self._metric_logging_function:
             return None
 
@@ -602,8 +806,15 @@ class Stream(metaclass=abc.ABCMeta):
             metric["tags"].update(extra_tags)
         self._metric_logging_function(f"INFO METRIC: {str(metric)}")
 
-    def _write_record_count_log(self, record_count, context: Optional[dict]) -> None:
-        """Emit a metric log. Optionally with appended tag info."""
+    def _write_record_count_log(
+        self, record_count: int, context: Optional[dict]
+    ) -> None:
+        """Emit a metric log. Optionally with appended tag info.
+
+        Args:
+            record_count: TODO
+            context: Stream partition or context dictionary.
+        """
         extra_tags = {} if not context else {"context": context}
         counter_metric: Dict[str, Any] = {
             "type": "counter",
@@ -620,6 +831,14 @@ class Stream(metaclass=abc.ABCMeta):
         context: Optional[dict],
         extra_tags: Optional[dict],
     ) -> None:
+        """TODO.
+
+        Args:
+            endpoint: TODO
+            response: TODO
+            context: Stream partition or context dictionary.
+            extra_tags: TODO
+        """
         request_duration_metric: Dict[str, Any] = {
             "type": "timer",
             "metric": "http_request_duration",
@@ -635,7 +854,15 @@ class Stream(metaclass=abc.ABCMeta):
             extra_tags["context"] = context
         self._write_metric_log(metric=request_duration_metric, extra_tags=extra_tags)
 
-    def _check_max_record_limit(self, record_count) -> None:
+    def _check_max_record_limit(self, record_count: int) -> None:
+        """TODO.
+
+        Args:
+            record_count: TODO.
+
+        Raises:
+            MaxRecordsLimitException: TODO.
+        """
         if (
             self._MAX_RECORDS_LIMIT is not None
             and record_count >= self._MAX_RECORDS_LIMIT
@@ -651,6 +878,9 @@ class Stream(metaclass=abc.ABCMeta):
         """Reset progress markers. If all=True, all state contexts will be set.
 
         This method is internal to the SDK and should not need to be overridden.
+
+        Args:
+            state: State object to promote progress markers with.
         """
         if state is None or state == {}:
             context: Optional[dict]
@@ -668,6 +898,9 @@ class Stream(metaclass=abc.ABCMeta):
         This method is internal to the SDK and should not need to be overridden.
 
         If all=True and the stream has children, child streams will also be finalized.
+
+        Args:
+            state: State object to promote progress markers with.
         """
         if state is None or state == {}:
             for child_stream in self.child_streams or []:
@@ -687,7 +920,14 @@ class Stream(metaclass=abc.ABCMeta):
     def _sync_records(  # noqa C901  # too complex
         self, context: Optional[dict] = None
     ) -> None:
-        """Sync records, emitting RECORD and STATE messages."""
+        """Sync records, emitting RECORD and STATE messages.
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Raises:
+            InvalidStreamSortException: TODO
+        """
         record_count = 0
         current_context: Optional[dict]
         context_list: Optional[List[dict]]
@@ -699,6 +939,7 @@ class Stream(metaclass=abc.ABCMeta):
             current_context = current_context or None
             state = self.get_context_state(current_context)
             state_partition_context = self._get_state_partition_context(current_context)
+            self._write_starting_replication_value(current_context)
             child_context: Optional[dict] = (
                 None if current_context is None else copy.copy(current_context)
             )
@@ -754,10 +995,13 @@ class Stream(metaclass=abc.ABCMeta):
     # Public methods ("final", not recommended to be overridden)
 
     @final
-    def sync(self, context: Optional[dict] = None):
+    def sync(self, context: Optional[dict] = None) -> None:
         """Sync this stream.
 
         This method is internal to the SDK and should not need to be overridden.
+
+        Args:
+            context: Stream partition or context dictionary.
         """
         msg = f"Beginning {self.replication_method.lower()} sync of '{self.name}'"
         if context:
@@ -774,7 +1018,7 @@ class Stream(metaclass=abc.ABCMeta):
         # Sync the records themselves:
         self._sync_records(context)
 
-    def _sync_children(self, child_context) -> None:
+    def _sync_children(self, child_context: dict) -> None:
         for child_stream in self.child_streams:
             if child_stream.selected or child_stream.has_selected_descendents:
                 child_stream.sync(context=child_context)
@@ -787,6 +1031,10 @@ class Stream(metaclass=abc.ABCMeta):
         Developers may override this method in order to introduce advanced catalog
         parsing, or to explicitly fail on advanced catalog customizations which
         are not supported by the tap.
+
+        Args:
+            catalog: Catalog object passed to the tap. Defines schema, primary and
+                replication keys, as well as selection metadata.
         """
         self._tap_input_catalog = catalog
 
@@ -798,7 +1046,14 @@ class Stream(metaclass=abc.ABCMeta):
                 self.forced_replication_method = catalog_entry.replication_method
 
     def _get_state_partition_context(self, context: Optional[dict]) -> Optional[Dict]:
-        """Override state handling if Stream.state_partitioning_keys is specified."""
+        """Override state handling if Stream.state_partitioning_keys is specified.
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Returns:
+            TODO
+        """
         if context is None:
             return None
 
@@ -814,6 +1069,17 @@ class Stream(metaclass=abc.ABCMeta):
 
         Developers may override this behavior to send specific information to child
         streams for context.
+
+        Args:
+            record: Individual record in the stream.
+            context: Stream partition or context dictionary.
+
+        Returns:
+            A dictionary with context values for a child stream.
+
+        Raises:
+            NotImplementedError: If the stream has children but this method is not
+                overriden.
         """
         if context is None:
             for child_stream in self.child_streams:
@@ -849,21 +1115,34 @@ class Stream(metaclass=abc.ABCMeta):
 
         This method takes an optional `context` argument, which can be safely ignored
         unless the stream is a child stream or requires partitioning.
-        More info: https://sdk.meltano.com/en/latest/partitioning.html
+        More info: :doc:`/partitioning`.
 
         Parent streams can optionally return a tuple, in which
         case the second item in the tuple being a `child_context` dictionary for the
         stream's `context`.
-        More info: https://sdk.meltano.com/en/latest/parent_streams.html
+        More info: :doc:`/parent_streams`
+
+        Args:
+            context: Stream partition or context dictionary.
         """
         pass
 
-    def post_process(self, row: dict, context: Optional[dict] = None) -> dict:
+    def post_process(self, row: dict, context: Optional[dict] = None) -> Optional[dict]:
         """As needed, append or transform raw data to match expected structure.
 
         Optional. This method gives developers an opportunity to "clean up" the results
         prior to returning records to the downstream tap - for instance: cleaning,
         renaming, or appending properties to the raw record result returned from the
         API.
+
+        Developers may also return `None` from this method to filter out
+        invalid or not-applicable records from the stream.
+
+        Args:
+            row: Individual record in the stream.
+            context: Stream partition or context dictionary.
+
+        Returns:
+            The resulting record dict, or `None` if the record should be excluded.
         """
         return row
