@@ -1,22 +1,23 @@
 """Abstract base class for API-type streams."""
 
 import abc
-import backoff
 import copy
 import logging
-import requests
-
 from datetime import datetime
-from typing import Any, Dict, Iterable, List, Optional, Union, cast
+from typing import Any, Callable, Dict, Iterable, List, Optional, Union, cast
 
+import backoff
+import requests
 from singer.schema import Schema
 
 from singer_sdk.authenticators import APIAuthenticatorBase, SimpleAuthenticator
+from singer_sdk.exceptions import FatalAPIError, RetriableAPIError
 from singer_sdk.helpers.jsonpath import extract_jsonpath
 from singer_sdk.plugin_base import PluginBase as TapBaseClass
 from singer_sdk.streams.core import Stream
 
 DEFAULT_PAGE_SIZE = 1000
+DEFAULT_REQUEST_TIMEOUT = 300  # 5 minutes
 
 
 class RESTStream(Stream, metaclass=abc.ABCMeta):
@@ -41,7 +42,7 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
     @property
     @abc.abstractmethod
     def url_base(self) -> str:
-        """Return the base url, e.g. 'https://api.mysite.com/v3/'."""
+        """Return the base url, e.g. ``https://api.mysite.com/v3/``."""
         pass
 
     def __init__(
@@ -50,8 +51,15 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
         name: Optional[str] = None,
         schema: Optional[Union[Dict[str, Any], Schema]] = None,
         path: Optional[str] = None,
-    ):
-        """Initialize the REST stream."""
+    ) -> None:
+        """Initialize the REST stream.
+
+        Args:
+            tap: Singer Tap this stream belongs to.
+            schema: JSON schema for records in this stream.
+            name: Name of this stream.
+            path: URL path for this entity stream.
+        """
         super().__init__(name=name, schema=schema, tap=tap)
         if path:
             self.path = path
@@ -62,7 +70,14 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
 
     @staticmethod
     def _url_encode(val: Union[str, datetime, bool, int, List[str]]) -> str:
-        """Encode the val argument as url-compatible string."""
+        """Encode the val argument as url-compatible string.
+
+        Args:
+            val: TODO
+
+        Returns:
+            TODO
+        """
         if isinstance(val, str):
             result = val.replace("/", "%2F")
         else:
@@ -70,9 +85,15 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
         return result
 
     def get_url(self, context: Optional[dict]) -> str:
-        """Return a URL, optionally targeted to a specific partition or context.
+        """Get stream entity URL.
 
         Developers override this method to perform dynamic URL generation.
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Returns:
+            A URL, optionally targeted to a specific partition or context.
         """
         url = "".join([self.url_base, self.path or ""])
         vals = copy.copy(dict(self.config))
@@ -87,47 +108,105 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
 
     @property
     def requests_session(self) -> requests.Session:
-        """Return the session object for HTTP requests."""
+        """Get requests session.
+
+        Returns:
+            The `requests.Session`_ object for HTTP requests.
+
+        .. _requests.Session:
+            https://docs.python-requests.org/en/latest/api/#request-sessions
+        """
         if not self._requests_session:
             self._requests_session = requests.Session()
         return self._requests_session
 
-    @backoff.on_exception(
-        backoff.expo,
-        (requests.exceptions.RequestException),
-        max_tries=5,
-        giveup=lambda e: e.response is not None and 400 <= e.response.status_code < 500,
-        factor=2,
-    )
-    def _request_with_backoff(
-        self, prepared_request, context: Optional[dict]
+    def validate_response(self, response: requests.Response) -> None:
+        """Validate HTTP response.
+
+        By default, checks for error status codes (>400) and raises a
+        :class:`singer_sdk.exceptions.FatalAPIError`.
+
+        Tap developers are encouraged to override this method if their APIs use HTTP
+        status codes in non-conventional ways, or if they communicate errors
+        differently (e.g. in the response body).
+
+        .. image:: ../images/200.png
+
+
+        In case an error is deemed transient and can be safely retried, then this
+        method should raise an :class:`singer_sdk.exceptions.RetriableAPIError`.
+
+        Args:
+            response: A `requests.Response`_ object.
+
+        Raises:
+            FatalAPIError: If the request is not retriable.
+            RetriableAPIError: If the request is retriable.
+
+        .. _requests.Response:
+            https://docs.python-requests.org/en/latest/api/#requests.Response
+        """
+        if 400 <= response.status_code < 500:
+            msg = (
+                f"{response.status_code} Client Error: "
+                f"{response.reason} for path: {self.path}"
+            )
+            raise FatalAPIError(msg)
+
+        elif 500 <= response.status_code < 600:
+            msg = (
+                f"{response.status_code} Server Error: "
+                f"{response.reason} for path: {self.path}"
+            )
+            raise RetriableAPIError(msg)
+
+    def request_decorator(self, func: Callable) -> Callable:
+        """Instantiate a decorator for handling request failures.
+
+        Developers may override this method to provide custom backoff or retry
+        handling.
+
+        Args:
+            func: Function to decorate.
+
+        Returns:
+            A decorated method.
+        """
+        decorator: Callable = backoff.on_exception(
+            backoff.expo,
+            (
+                RetriableAPIError,
+                requests.exceptions.ReadTimeout,
+            ),
+            max_tries=5,
+            factor=2,
+        )(func)
+        return decorator
+
+    def _request(
+        self, prepared_request: requests.PreparedRequest, context: Optional[dict]
     ) -> requests.Response:
-        response = self.requests_session.send(prepared_request)
+        """TODO.
+
+        Args:
+            prepared_request: TODO
+            context: Stream partition or context dictionary.
+
+        Returns:
+            TODO
+        """
+        response = self.requests_session.send(prepared_request, timeout=self.timeout)
         if self._LOG_REQUEST_METRICS:
             extra_tags = {}
             if self._LOG_REQUEST_METRIC_URLS:
-                extra_tags["url"] = cast(str, prepared_request.path_url)
+                extra_tags["url"] = prepared_request.path_url
             self._write_request_duration_log(
                 endpoint=self.path,
                 response=response,
                 context=context,
                 extra_tags=extra_tags,
             )
-        if response.status_code in [401, 403]:
-            self.logger.info("Failed request for {}".format(prepared_request.url))
-            self.logger.info(
-                f"Reason: {response.status_code} - {str(response.content)}"
-            )
-            raise RuntimeError(
-                "Requested resource was unauthorized, forbidden, or not found."
-            )
-        elif response.status_code >= 400:
-            raise RuntimeError(
-                f"Error making request to API: {prepared_request.url} "
-                f"[{response.status_code} - {str(response.content)}]".replace(
-                    "\\n", "\n"
-                )
-            )
+        self.validate_response(response)
         logging.debug("Response received successfully.")
         return response
 
@@ -137,6 +216,14 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
         """Return a dictionary of values to be used in URL parameterization.
 
         If paging is supported, developers may override with specific paging logic.
+
+        Args:
+            context: Stream partition or context dictionary.
+            next_page_token: Token, page number or any request argument to request the
+                next page of data.
+
+        Returns:
+            Dictionary of URL query parameters to use in the request.
         """
         return {}
 
@@ -148,6 +235,15 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
         If partitioning is supported, the `context` object will contain the partition
         definitions. Pagination information can be parsed from `next_page_token` if
         `next_page_token` is not None.
+
+        Args:
+            context: Stream partition or context dictionary.
+            next_page_token: Token, page number or any request argument to request the
+                next page of data.
+
+        Returns:
+            Build a request with the stream's URL, path, query parameters,
+            HTTP headers and authenticator.
         """
         http_method = self.rest_method
         url: str = self.get_url(context)
@@ -169,7 +265,7 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
                     params=params,
                     headers=headers,
                     json=request_data,
-                )
+                ),
             ),
         )
         return request
@@ -178,14 +274,26 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
         """Request records from REST endpoint(s), returning response records.
 
         If pagination is detected, pages will be recursed automatically.
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Yields:
+            An item for every record in the response.
+
+        Raises:
+            RuntimeError: If a loop in pagination is detected. That is, when two
+                consecutive pagination tokens are identical.
         """
         next_page_token: Any = None
         finished = False
+        decorated_request = self.request_decorator(self._request)
+
         while not finished:
             prepared_request = self.prepare_request(
                 context, next_page_token=next_page_token
             )
-            resp = self._request_with_backoff(prepared_request, context)
+            resp = decorated_request(prepared_request, context)
             for row in self.parse_response(resp):
                 yield row
             previous_token = copy.deepcopy(next_page_token)
@@ -212,13 +320,32 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
         Developers may override this method if the API requires a custom payload along
         with the request. (This is generally not required for APIs which use the
         HTTP 'GET' method.)
+
+        Args:
+            context: Stream partition or context dictionary.
+            next_page_token: Token, page number or any request argument to request the
+                next page of data.
+
+        Returns:
+            Dictionary with the body to use for the request.
         """
         return None
 
     def get_next_page_token(
         self, response: requests.Response, previous_token: Optional[Any]
     ) -> Any:
-        """Return token identifying next page or None if all records have been read."""
+        """Return token identifying next page or None if all records have been read.
+
+        Args:
+            response: A raw `requests.Response`_ object.
+            previous_token: Previous pagination reference.
+
+        Returns:
+            Reference value to retrieve next page.
+
+        .. _requests.Response:
+            https://docs.python-requests.org/en/latest/api/#requests.Response
+        """
         if self.next_page_token_jsonpath:
             all_matches = extract_jsonpath(
                 self.next_page_token_jsonpath, response.json()
@@ -236,25 +363,58 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
 
         If an authenticator is also specified, the authenticator's headers will be
         combined with `http_headers` when making HTTP requests.
+
+        Returns:
+            Dictionary of HTTP headers to use as a base for every request.
         """
         result = self._http_headers
         if "user_agent" in self.config:
             result["User-Agent"] = self.config.get("user_agent")
         return result
 
+    @property
+    def timeout(self) -> int:
+        """Return the request timeout limit in seconds.
+
+        The default timeout is 300 seconds, or as defined by DEFAULT_REQUEST_TIMEOUT.
+
+        Returns:
+            The request timeout limit as number of seconds.
+        """
+        return DEFAULT_REQUEST_TIMEOUT
+
     # Records iterator
 
     def get_records(self, context: Optional[dict]) -> Iterable[Dict[str, Any]]:
         """Return a generator of row-type dictionary objects.
 
         Each row emitted should be a dictionary of property names to their values.
+
+        Args:
+            context: Stream partition or context dictionary.
+
+        Yields:
+            One item per (possibly processed) record in the API.
         """
-        for row in self.request_records(context):
-            row = self.post_process(row, context)
-            yield row
+        for record in self.request_records(context):
+            transformed_record = self.post_process(record, context)
+            if transformed_record is None:
+                # Record filtered out during post_process()
+                continue
+            yield transformed_record
 
     def parse_response(self, response: requests.Response) -> Iterable[dict]:
-        """Parse the response and return an iterator of result rows."""
+        """Parse the response and return an iterator of result rows.
+
+        Args:
+            response: A raw `requests.Response`_ object.
+
+        Yields:
+            One item for every item found in the response.
+
+        .. _requests.Response:
+            https://docs.python-requests.org/en/latest/api/#requests.Response
+        """
         yield from extract_jsonpath(self.records_jsonpath, input=response.json())
 
     # Abstract methods:
@@ -265,5 +425,9 @@ class RESTStream(Stream, metaclass=abc.ABCMeta):
 
         If an authenticator is not specified, REST-based taps will simply pass
         `http_headers` as defined in the stream class.
+
+        Returns:
+            Authenticator instance that will be used to authenticate all outgoing
+            requests.
         """
         return SimpleAuthenticator(stream=self)
