@@ -2,27 +2,40 @@
 
 import abc
 import json
-from pathlib import PurePath, Path
-from singer_sdk.mapper import PluginMapper
-from typing import Any, List, Optional, Dict, Tuple, Type, Union, cast
+from enum import Enum
+from pathlib import Path, PurePath
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
 
 import click
 
+from singer_sdk.cli import common_options
+from singer_sdk.exceptions import MaxRecordsLimitException
+from singer_sdk.helpers import _state
 from singer_sdk.helpers._classproperty import classproperty
 from singer_sdk.helpers._compat import final
-from singer_sdk.helpers._util import read_json_file
 from singer_sdk.helpers._singer import Catalog
 from singer_sdk.helpers._state import write_stream_state
-from singer_sdk.plugin_base import PluginBase
-from singer_sdk.streams.core import Stream
-from singer_sdk.exceptions import (
-    MaxRecordsLimitException,
+from singer_sdk.helpers._util import read_json_file
+from singer_sdk.helpers.capabilities import (
+    CapabilitiesEnum,
+    PluginCapabilities,
+    TapCapabilities,
 )
-from singer_sdk.helpers import _state
+from singer_sdk.mapper import PluginMapper
+from singer_sdk.plugin_base import PluginBase
+from singer_sdk.streams import SQLStream, Stream
 
 STREAM_MAPS_CONFIG = "stream_maps"
 
 
+class CliTestOptionValue(Enum):
+    """Values for CLI option --test."""
+
+    All = "all"
+    Schema = "schema"
+    Disabled = False
+
+
 class Tap(PluginBase, metaclass=abc.ABCMeta):
     """Abstract base class for taps.
 
@@ -38,9 +51,25 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
         catalog: Union[PurePath, str, dict, None] = None,
         state: Union[PurePath, str, dict, None] = None,
         parse_env_config: bool = False,
+        validate_config: bool = True,
     ) -> None:
-        """Initialize the tap."""
-        super().__init__(config=config, parse_env_config=parse_env_config)
+        """Initialize the tap.
+
+        Args:
+            config: Tap configuration. Can be a dictionary, a single path to a
+                configuration file, or a list of paths to multiple configuration
+                files.
+            catalog: Tap catalog. Can be a dictionary or a path to the catalog file.
+            state: Tap state. Can be dictionary or a path to the state file.
+            parse_env_config: Whether to look for configuration values in environment
+                variables.
+            validate_config: True to require validation of config settings.
+        """
+        super().__init__(
+            config=config,
+            parse_env_config=parse_env_config,
+            validate_config=validate_config,
+        )
 
         # Declare private members
         self._streams: Optional[Dict[str, Stream]] = None
@@ -75,9 +104,12 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
 
     @property
     def streams(self) -> Dict[str, Stream]:
-        """Return a list of streams, using discovery or a provided catalog.
+        """Get streams discovered or catalogued for this tap.
 
         Results will be cached after first execution.
+
+        Returns:
+            A mapping of names to streams, using discovery or a provided catalog.
         """
         input_catalog = self.input_catalog
 
@@ -91,26 +123,56 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
 
     @property
     def state(self) -> dict:
-        """Return a state dict."""
+        """Get tap state.
+
+        Returns:
+            The tap's state dictionary
+
+        Raises:
+            RuntimeError: If state has not been initialized.
+        """
         if self._state is None:
             raise RuntimeError("Could not read from uninitialized state.")
         return self._state
 
     @property
     def input_catalog(self) -> Optional[Catalog]:
-        """Return the catalog dictionary input, or None if not provided."""
+        """Get the catalog passed to the tap.
+
+        Returns:
+            Catalog dictionary input, or None if not provided.
+        """
         return self._input_catalog
 
     @classproperty
-    def capabilities(self) -> List[str]:
-        """Return a list of supported capabilities."""
-        return ["sync", "catalog", "state", "discover"]
+    def capabilities(self) -> List[CapabilitiesEnum]:
+        """Get tap capabilities.
+
+        Returns:
+            A list of capabilities supported by this tap.
+        """
+        return [
+            TapCapabilities.CATALOG,
+            TapCapabilities.STATE,
+            TapCapabilities.DISCOVER,
+            PluginCapabilities.ABOUT,
+            PluginCapabilities.STREAM_MAPS,
+            PluginCapabilities.FLATTENING,
+        ]
 
     # Connection test:
 
     @final
     def run_connection_test(self) -> bool:
-        """Run connection test and return True if successful."""
+        """Run connection test.
+
+        Returns:
+            True if the test succeeded.
+        """
+        for stream in self.streams.values():
+            # Initialize streams' record limits before beginning the sync test.
+            stream._MAX_RECORDS_LIMIT = 1
+
         for stream in self.streams.values():
             if stream.parent_stream_type:
                 self.logger.debug(
@@ -119,42 +181,70 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
                     "Skipping direct invocation."
                 )
                 continue
-
-            stream._MAX_RECORDS_LIMIT = 1 if stream.child_streams else 0
             try:
                 stream.sync()
             except MaxRecordsLimitException:
                 pass
         return True
 
+    @final
+    def write_schemas(self) -> None:
+        """Write a SCHEMA message for all known streams to STDOUT."""
+        for stream in self.streams.values():
+            stream._write_schema_message()
+
     # Stream detection:
 
     def run_discovery(self) -> str:
-        """Write the catalog json to STDOUT and return the same as a string."""
+        """Write the catalog json to STDOUT and return as a string.
+
+        Returns:
+            The catalog as a string of JSON.
+        """
         catalog_text = self.catalog_json_text
         print(catalog_text)
         return catalog_text
 
     @property
     def catalog_dict(self) -> dict:
-        """Return the tap's catalog as a dict."""
+        """Get catalog dictionary.
+
+        Returns:
+            The tap's catalog as a dict
+        """
         return cast(dict, self._singer_catalog.to_dict())
 
     @property
     def catalog_json_text(self) -> str:
-        """Return the tap's catalog as formatted json text."""
+        """Get catalog JSON.
+
+        Returns:
+            The tap's catalog as formatted JSON text.
+        """
         return json.dumps(self.catalog_dict, indent=2)
 
     @property
     def _singer_catalog(self) -> Catalog:
-        """Return a Catalog object."""
+        """Return a Catalog object.
+
+        Returns:
+            :class:`singer_sdk.helpers._singer.Catalog`.
+        """
         return Catalog(
             (stream.tap_stream_id, stream._singer_catalog_entry)
             for stream in self.streams.values()
         )
 
     def discover_streams(self) -> List[Stream]:
-        """Initialize all available streams and return them as a list."""
+        """Initialize all available streams and return them as a list.
+
+        Return:
+            List of discovered Stream objects.
+
+        Raises:
+            NotImplementedError: If the tap implementation does not override this
+                method.
+        """
         raise NotImplementedError(
             f"Tap '{self.name}' does not support discovery. "
             "Please set the '--catalog' command line argument and try again."
@@ -166,6 +256,9 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
 
         Return the output of `self.discover_streams()` to enumerate
         discovered streams.
+
+        Returns:
+            A list of discovered streams, ordered by name.
         """
         # Build the parent-child dependency DAG
 
@@ -203,6 +296,13 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
         Override this method to perform validation and backwards-compatibility patches
         on self.state. If overriding, we recommend first running
         `super().load_state(state)` to ensure compatibility with the SDK.
+
+        Args:
+            state: Initialize the tap'ss state with this value.
+
+        Raises:
+            ValueError: If the tap's own state is None, meaning it has not been
+                initialized.
         """
         if self.state is None:
             raise ValueError("Cannot write to uninitialized state dictionary.")
@@ -220,7 +320,7 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
 
     def _reset_state_progress_markers(self) -> None:
         """Clear prior jobs' progress markers at beginning of sync."""
-        for stream_name, state in self.state.get("bookmarks", {}).items():
+        for _, state in self.state.get("bookmarks", {}).items():
             _state.reset_state_progress_markers(state)
             for partition_state in state.get("partitions", []):
                 _state.reset_state_progress_markers(partition_state)
@@ -244,7 +344,7 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
     # Sync methods
 
     @final
-    def sync_all(self):
+    def sync_all(self) -> None:
         """Sync all streams."""
         self._reset_state_progress_markers()
         self._set_compatible_replication_methods()
@@ -268,19 +368,17 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
     # Command Line Execution
 
     @classproperty
-    def cli(cls):
-        """Execute standard CLI handler for taps."""
+    def cli(cls) -> Callable:
+        """Execute standard CLI handler for taps.
 
-        @click.option(
-            "--version",
-            is_flag=True,
-            help="Display the package version.",
-        )
-        @click.option(
-            "--about",
-            is_flag=True,
-            help="Display package metadata and settings.",
-        )
+        Returns:
+            A callable CLI object.
+        """
+
+        @common_options.PLUGIN_VERSION
+        @common_options.PLUGIN_ABOUT
+        @common_options.PLUGIN_ABOUT_FORMAT
+        @common_options.PLUGIN_CONFIG
         @click.option(
             "--discover",
             is_flag=True,
@@ -288,21 +386,14 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
         )
         @click.option(
             "--test",
-            is_flag=True,
-            help="Test connectivity by syncing a single record and exiting.",
-        )
-        @click.option(
-            "--format",
-            help="Specify output style for --about",
-            type=click.Choice(["json"], case_sensitive=False),
-            default=None,
-        )
-        @click.option(
-            "--config",
-            multiple=True,
-            help="Configuration file location or 'ENV' to use environment variables.",
-            type=click.STRING,
-            default=(),
+            is_flag=False,
+            flag_value=CliTestOptionValue.All.value,
+            default=CliTestOptionValue.Disabled,
+            help=(
+                "Use --test to sync a single record for each stream. "
+                + "Use --test=schema to test schema output without syncing "
+                + "records."
+            ),
         )
         @click.option(
             "--catalog",
@@ -322,22 +413,39 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
             version: bool = False,
             about: bool = False,
             discover: bool = False,
-            test: bool = False,
+            test: CliTestOptionValue = CliTestOptionValue.Disabled,
             config: Tuple[str, ...] = (),
             state: str = None,
             catalog: str = None,
             format: str = None,
         ) -> None:
-            """Handle command line execution."""
+            """Handle command line execution.
+
+            Args:
+                version: Display the package version.
+                about: Display package metadata and settings.
+                discover: Run the tap in discovery mode.
+                test: Test connectivity by syncing a single record and exiting.
+                format: Specify output style for `--about`.
+                config: Configuration file location or 'ENV' to use environment
+                    variables. Accepts multiple inputs as a tuple.
+                catalog: Use a Singer catalog file with the tap.",
+                state: Use a bookmarks file for incremental replication.
+
+            Raises:
+                FileNotFoundError: If the config file does not exist.
+            """
             if version:
                 cls.print_version()
                 return
 
-            if about:
-                cls.print_about(format)
-                return
+            if not about:
+                cls.print_version(print_fn=cls.logger.info)
 
-            cls.print_version(print_fn=cls.logger.info)
+            validate_config: bool = True
+            if about or discover:
+                # Don't abort on validation failures
+                validate_config = False
 
             parse_env_config = False
             config_files: List[PurePath] = []
@@ -361,14 +469,90 @@ class Tap(PluginBase, metaclass=abc.ABCMeta):
                 state=state,
                 catalog=catalog,
                 parse_env_config=parse_env_config,
+                validate_config=validate_config,
             )
-            if discover:
+            if about:
+                tap.print_about(format)
+            elif discover:
                 tap.run_discovery()
-                if test:
+                if test == CliTestOptionValue.All.value:
                     tap.run_connection_test()
-            elif test:
+            elif test == CliTestOptionValue.All.value:
                 tap.run_connection_test()
+            elif test == CliTestOptionValue.Schema.value:
+                tap.write_schemas()
             else:
                 tap.sync_all()
 
         return cli
+
+
+class SQLTap(Tap):
+    """A specialized Tap for extracting from SQL streams."""
+
+    # Stream class used to initialize new SQL streams from their catalog declarations.
+    default_stream_class: Type[SQLStream]
+
+    def __init__(
+        self,
+        config: Optional[Union[dict, PurePath, str, List[Union[PurePath, str]]]] = None,
+        catalog: Union[PurePath, str, dict, None] = None,
+        state: Union[PurePath, str, dict, None] = None,
+        parse_env_config: bool = False,
+        validate_config: bool = True,
+    ) -> None:
+        """Initialize the SQL tap.
+
+        The SQLTap initializer additionally creates a cache variable for _catalog_dict.
+
+        Args:
+            config: Tap configuration. Can be a dictionary, a single path to a
+                configuration file, or a list of paths to multiple configuration
+                files.
+            catalog: Tap catalog. Can be a dictionary or a path to the catalog file.
+            state: Tap state. Can be dictionary or a path to the state file.
+            parse_env_config: Whether to look for configuration values in environment
+                variables.
+            validate_config: True to require validation of config settings.
+        """
+        self._catalog_dict: Optional[dict] = None
+        super().__init__(
+            config=config,
+            catalog=catalog,
+            state=state,
+            parse_env_config=parse_env_config,
+            validate_config=validate_config,
+        )
+
+    @property
+    def catalog_dict(self) -> dict:
+        """Get catalog dictionary.
+
+        Returns:
+            The tap's catalog as a dict
+        """
+        if self._catalog_dict:
+            return self._catalog_dict
+
+        if self.input_catalog:
+            return self.input_catalog.to_dict()
+
+        connector = self.default_stream_class.connector_class(dict(self.config))
+
+        result: Dict[str, List[dict]] = {"streams": []}
+        result["streams"].extend(connector.discover_catalog_entries())
+
+        self._catalog_dict = result
+        return self._catalog_dict
+
+    def discover_streams(self) -> List[Stream]:
+        """Initialize all available streams and return them as a list.
+
+        Returns:
+            List of discovered Stream objects.
+        """
+        result: List[Stream] = []
+        for catalog_entry in self.catalog_dict["streams"]:
+            result.append(self.default_stream_class(self, catalog_entry))
+
+        return result
